# Copyright (c) 2001-2007, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.

package PAPI::Cookie_Handler;


use PAPI::ApachePoA;
use PAPI::RedirectGPoA; 
use MIME::Base64 qw/decode_base64 encode_base64/;
use Crypt::OpenSSL::RSA;
use constant PADDINGSIZE => 11;
use URI::Escape;
use Digest::MD5 qw(md5_base64);
use Apache2::Response;
use Apache2::SubRequest;
use Apache2::Const qw(REDIRECT OK);
use Apache2::RequestIO;

sub handler
{
	my $r=shift;
	my $poa=new PAPI::ApachePoA;

	$poa->set_parameters($r);
	my %fields= $poa->stringArgsToHash($r->args());

	if (($fields{'ACTION'} ne "LOGIN") && ($fields{'ACTION'} ne "TEST") && ($fields{'ACTION'} ne "LOGOUT") && ($fields{'ACTION'} ne "CHECK") && ($fields{'ACTION'} ne "CHECKED")){
		$poa->warning("Signed URL without valid ACTION field:$fields{'ACTION'}");
		return ERROR;	
	}	
	if ($fields{'DATA'} eq ""){
		$poa->warning("Signed URL without DATA field");
		return ERROR;
	}
	
	### PROCESSING ACTIONS #############################

	if ($fields{'ACTION'} eq "TEST"){
		my $rtv="reject";
		my ($Hcook,$Lcook)=$poa->get_cookies();
		my ($ret,$code)=$poa->test_Lcook($Lcook);
		if ($ret>0) { $rtv="accept"; }
		elsif (defined $Hcook) {
			my ($valid_date,$temp,$new_crc,$gpoa_vdate);
			($code,$valid_date,$temp,$new_crc,$gpoa_vdate)=$poa->test_Hcook($Hcook,0);
			if ($valid_date>0) { $rtv="accept"; }
		}
		if ($rtv eq "accept") {
			if (defined $fields{'AURL'}) { return redirect_URL($r, uri_unescape($fields{'AURL'})); }
			else { return send_file($r,$poa->{'Accept_File'},$poa); }
		}
		else {
			if (defined $fields{'RURL'}) { return redirect_URL($r, uri_unescape($fields{'RURL'})); }
			else { return send_file($r,$poa->{'Reject_File'},$poa); }
		}
	}
	if ($fields{'ACTION'} eq "LOGOUT"){
		my $rtv="reject";
		my ($Hcook,$Lcook)=$poa->get_cookies();
		my ($ret,$code)=$poa->test_Lcook($Lcook);
		if ($ret>0) { $rtv="accept"; }
		elsif (defined $Hcook) {
			my ($valid_date,$temp,$new_crc,$gpoa_vdate);
			($code,$valid_date,$temp,$new_crc,$gpoa_vdate)=$poa->test_Hcook($Hcook,0);
			if ($valid_date>0) { $rtv="accept"; }
		}	
		if ($rtv eq "accept") {
			$poa->accesslog("Logging out $code");
			($Hcook,$Lcook)=$poa->gen_logout_cookies($code);
			if (defined $fields{'AURL'}) {
				return redirect_with_cookies($r,uri_unescape($fields{'AURL'}),$Hcook,$Lcook);
			}
			else {
				return send_file_with_cookies($r,$Hcook,$Lcook,$poa->{'Accept_File'},$poa);
			}
		}
		else {	
			$poa->warning("Error trying to LOGOUT");
			if (defined $fields{'RURL'}) {
				return redirect_URL($r, uri_unescape($fields{'RURL'}));
			}
			else {
				return send_file($r,$poa->{'Reject_File'},$poa);
			}
		}
	}
	if ($fields{'ACTION'} eq "LOGIN"){
		my ($code,$valid_date)=$poa->test_URL($fields{'AS'},$fields{'DATA'});
		if ($valid_date == 0){
			$poa->warning("Error validating URL from AS $fields{'AS'}");
			if (defined $fields{'RURL'}) {
				return redirect_URL($r, uri_unescape($fields{'RURL'}));
			}
			else {
				return send_file($r,$poa->{'Reject_File'},$poa);
			}
		}
		my $Lcook=$poa->gen_Lcook($code);
		my $Hcook=undef;
		if ($poa->{'useHcook'}) { 
			$Hcook=$poa->gen_Hcook($code,$valid_date,0);
			$r->pnotes('PAPIregisterKey' => $poa->{'registerKey'});
			$r->pnotes('PAPIregisterVal' => $poa->{'registerVal'});
		}
		if (defined $fields{'AURL'}) {
			return redirect_with_cookies($r,uri_unescape($fields{'AURL'}),$Hcook,$Lcook);
		}
		else {
			return send_file_with_cookies($r,$Hcook,$Lcook,$poa->{'Accept_File'},$poa);
		}
	}
	if ($fields{'ACTION'} eq "CHECK"){
		my ($Hcook,$Lcook)=$poa->get_cookies();
		my ($lrt,$code,$valid_date,$temp,$new_crc,$gpoa_vdate);
		my $now=time;
		($lrt,$code)=$poa->test_Lcook($Lcook);
		if ($lrt == 0) {
			if (!defined $Hcook) {
				if ($poa->{GPoA_URL} ne "") { return PAPI::RedirectGPoA::handler ($r, $poa); }
				else {
					$poa->warning("Error testing authZ tokens at GPoA for $fields{'URL'}");
					return redirect_ERROR_URL($r,$poa,$fields{'URL'},$fields{'DATA'});
				}
			}
			else {
				($code,$valid_date,$temp,$new_crc,$gpoa_vdate)=$poa->test_Hcook($Hcook,1);
				if ($valid_date <= 0){
					if ($valid_date == 0 && $poa->{GPoA_URL} ne ""){ return PAPI::RedirectGPoA::handler ($r, $poa); }
					else {
						$poa->warning("Error testing authZ tokens at GPoA for $fields{'URL'}");
						return redirect_ERROR_URL($r,$poa,$fields{'URL'},$fields{'DATA'});
					}
				}
			}					
		}
		else {
			$valid_date=1;
			$gpoa_vdate=$now+$poa->{'Lcook_Timeout'};
		}
		my $poaURL=uri_unescape($fields{'URL'});
		my $opoa=$poaURL;
		if (defined $fields{'PAPIOPOA'}) { $opoa=uri_unescape($fields{'PAPIOPOA'}); }
		my @uasc=split /@/,$code;
		my $asid=pop @uasc;
		my $uas=join '@',@uasc;
		for my $poarw (@{$poa->{PoARw}}) {
			if ($opoa =~ /$poarw->{poare}/) {
				my $evs = "\$uas =~ s($poarw->{regex})($poarw->{value})gsx";
				eval $evs;
			}
		}
		my $newcode;
		if (defined $poa->{GPoAHashUser} && $poa->{GPoAHashUser} eq "1") { 
			$newcode=md5_base64($uas).'@'.$asid; 
		}
		else { $newcode=$uas.'@'.$asid; }
		
		# Commented-out 20070926 by Dani.
		# The new code takes care of data longer than key size
		# my $pkey = Crypt::OpenSSL::RSA->new_private_key(read_key_file($poa->{'GPoA_Priv_Key'}));
		# $pkey->use_pkcs1_padding();
		# my $safe = encode_base64($pkey->private_encrypt("$newcode:$gpoa_vdate:$now:$fields{'DATA'}"));
		my $safe = PAPI::PoA::rsaEncrypt("$newcode:$gpoa_vdate:$now:$fields{'DATA'}",$poa->{'GPoA_Priv_Key'});
		
		if ($poaURL =~ /\?/) { $poaURL .= "&ACTION=CHECKED"; }
		else { $poaURL .= "?ACTION=CHECKED"; }
		$poaURL .= "&DATA=" . uri_escape("$safe",'\W');
		if($valid_date > 1){
			my $Lcook=$poa->gen_Lcook($code);
			$poa->debug("Main: New Lcook: $Lcook") if ($poa->{'Debug'});
			my $Hcook=undef;
			if ($poa->{'useHcook'}) {
				$Hcook=$poa->gen_Hcook($code,$valid_date,$temp,$new_crc);
				$poa->debug("Main: New Hcook: $Hcook") if ($poa->{'Debug'});
			}
			return redirect_with_cookies($r,$poaURL,$Hcook,$Lcook);
		}
		else{ return redirect_URL($r,$poaURL); }
	}
	if ($fields{'ACTION'} eq "CHECKED"){
		### Answer from a parent GPoA ####

		my $url=$r->uri;
		### Check from an answer of an AS to an ATTREQ ###
		my $isas=$fields{'AS'};
		my ($req_id,$code,$valid_date)=$poa->test_GPoA_URL($fields{'DATA'},$isas);
		my $data_req=$poa->load_request($req_id);
# If no request found, redirect it to the root location of the (G)PoA.
# Although this should not happen in normal behavior, it is a fallback
# to avoid redirection loops
		if (!defined $data_req) {
			$poa->error("Invalid requestId received from GPoA/AS");
			return redirect_URL($r, "http://".$r->hostname.$poa->{'Loc'});
		}
		my %tfields;
		for my $ti (split(/&/,$data_req->{args})){ if ($ti =~ /(.*)=(.*)/){ $tfields{$1}=$2; } }
		if ($valid_date == 0){
			$poa->warning("Invalid GPoA/AS Answer for $url");
			return redirect_ERROR_URL($r,$poa,$tfields{'URL'},$tfields{'DATA'});
		}
		my $poaURL=uri_unescape($tfields{'URL'});
		my $opoa=$poaURL;
		if (defined $fields{'PAPIOPOA'}) { $opoa=uri_unescape($fields{'PAPIOPOA'}); }
		my @uasc=split /@/,$code;
		my $asid=pop @uasc;
		my $uas=join '@',@uasc;
		for my $poarw (@{$poa->{PoARw}}) {
			if ($opoa =~ /$poarw->{poare}/) {
				my $evs = "\$uas =~ s($poarw->{regex})($poarw->{value})gsx";
				eval $evs;
			}
		}
		my $newcode;
		if (defined $poa->{GPoAHashUser} && $poa->{GPoAHashUser} eq "1") { 
			$newcode=md5_base64($uas).'@'.$asid; 
		}
		else { $newcode=$uas.'@'.$asid; }
		my $now=time;
		
		# Commented-out 20070926 by Dani.
		# The new code takes care of data longer than key size
		# my $pkey = Crypt::OpenSSL::RSA->new_private_key(read_key_file($poa->{'GPoA_Priv_Key'}));
		# $pkey->use_pkcs1_padding();
		# my $safe = encode_base64($pkey->private_encrypt("$newcode:$valid_date:$now:$tfields{'DATA'}"));
		
		my $safe = PAPI::PoA::rsaEncrypt("$newcode:$valid_date:$now:$tfields{'DATA'}",$poa->{'GPoA_Priv_Key'});
		if ($poaURL =~ /\?/) { $poaURL .= "&ACTION=CHECKED"; }
		else { $poaURL .= "?ACTION=CHECKED"; }
		$poaURL .= "&DATA=" . uri_escape("$safe",'\W');
		if($valid_date > 1){
			my $Lcook=$poa->gen_Lcook($code);
			$poa->debug("Main: New Lcook: $Lcook") if ($poa->{'Debug'});
			my $Hcook=undef;
			if ($poa->{'useHcook'}) {
				$Hcook=$poa->gen_Hcook($code,$valid_date,1);
				$poa->debug("Main: New Hcook: $Hcook") if ($poa->{'Debug'});
			}
			return redirect_with_cookies($r,$poaURL,$Hcook,$Lcook);
		}
		else{ return redirect_URL($r,$poaURL); }
	}
}

sub send_file
{
	my $r=shift;
	my $file=shift;
	my $poa=shift;
	
	my $sr=$r->lookup_file($file);
#	$sr->no_cache(1);
	my $res=$sr->run();
	$poa->error("Cookie_Handler: Exception handling $file") if ($res != OK);
	return $res;
}

sub send_file_with_cookies
{
	my $r=shift;
	my $Hcook=shift;
	my $Lcook=shift;
	my $file=shift;
	my $poa=shift;
	
	my $cty=$r->lookup_file($file)->content_type();
	my $no_bytes = (stat ($file))[7];

	$r->headers_out->set('Contennt-type' => $cty);
	if (defined $Hcook) { $r->headers_out->add('Set-cookie' => $Hcook); }
	$r->headers_out->add('Set-cookie' => $Lcook);
	$r->headers_out->set('Content-length' => $no_bytes);
	$r->status(Apache2::Const::OK); 	

 	$rc = $r->sendfile($file);
 	if ($rc != APR::Const::SUCCESS){
 		$poa->error("Cookie_handler: Unable to send $file: $!");
 		return $rc;
 	}
 	$r->rflush;
 	return Apache2::Const::OK;
}

sub redirect_with_cookies
{
	my $r=shift;
	my $uri=shift;
	my $Hcook=shift;
	my $Lcook=shift;

	my $log = $r->log();
	$log->info("Cookie_Handler::Redirect_with_cookies URI: $uri");

    if (defined $Hcook) { $r->err_headers_out->add('Set-Cookie' => $Hcook); }
    $r->err_headers_out->add('Set-Cookie' => $Lcook);
    $r->err_headers_out->set(Location => $uri);    
    $r->status(REDIRECT);
	return Apache2::Const::REDIRECT;
}

sub redirect_ERROR_URL
{
	my $r=shift;
	my $poa=shift;
	my $url=shift;
	my $data=shift;

	# Commented-out 20070926 by Dani.
	# The new code takes care of data longer than key size
	# my $pkey = Crypt::OpenSSL::RSA->new_private_key(read_key_file());
	# $pkey->use_pkcs1_padding();
	# my $safe = encode_base64($pkey->private_encrypt());
	my $safe = PAPI::PoA::rsaEncrypt("ERROR:0:0:$data",$poa->{'GPoA_Priv_Key'});

	my $poaURL = uri_unescape($url);
	if ($poaURL =~ /\?/) { $poaURL .= "&ACTION=CHECKED"; }
	else { $poaURL .= "?ACTION=CHECKED"; }
	$poaURL .= "&DATA=" . uri_escape("$safe",'\W');
	return redirect_URL($r,$poaURL);
}
 
sub redirect_URL
{
	my $r=shift;
	my $uri=shift;
	
	my $log = $r->log();
	$log->info("Cookie_Handler::Redirect_URL URI: $uri");

    $r->err_headers_out->set(Location => $uri);
    $r->status(REDIRECT);
	return Apache2::Const::REDIRECT;
}

sub read_key_file {
    my $file = shift;
    my $key_string;
    
    open(PRIV,$file) || die "$file: $!";
    read(PRIV,$key_string,-s PRIV); # Suck in the whole file
    close(PRIV);
    
    return $key_string;
}

=pod
sub rsaEncrypt {
	my $datarg = shift;
	my $keyFile = shift;
	
	# Get the byte size of data string 
	our $inputSize = 0;
	{
		use bytes;
		$inputSize = length($datarg);
	}

	# Generate a private key for encrypt and set the padding
	my $pkey = Crypt::OpenSSL::RSA->new_private_key(read_key_file($keyFile));
	$pkey->use_pkcs1_padding();

	# Get the output block maximun size
	### pkcs1_padding or sslv23_padding at most 11 bytes less than this size.
	# Get the size of block to encrypt and total number of blocks
	my $outputBlockSize = $pkey->size();
	my $inputBlockSize = $outputBlockSize - PADDINGSIZE;
	my $numBlocks = int($inputSize / $inputBlockSize);
	
	# Start encrypt.
	my $blockCount = 0;
	my @cryptBuffer;

	while ($blockCount < $numBlocks){
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($datarg, $index, $inputBlockSize);
		$cryptBuffer[$blockCount] = $pkey->private_encrypt($block);
		$blockCount++;
	}

	if ( ($inputSize % $inputBlockSize) != 0 ){
		$numBlocks++;
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($datarg, $index, $inputBlockSize);
		$cryptBuffer[$blockCount] = $pkey->private_encrypt($block);
	}

	# Now join the array with the encrypted block strings and return
	return encode_base64(join('',@cryptBuffer));
}

#
###
#

sub rsaDecrypt{
	my $cryptData = shift;
	my $keyFile = shift;

	# Decode the base64 input string
	$cryptData = decode_base64($cryptData);

	# Get the byte size of the data string 
	our $inputSize = 0;
	{
		use bytes;
		$inputSize = length($cryptData);
	}

	# Get the output block maximun size
	## PADDINGSIZE is not used to decrypt
	# Get the size of the encrypted block and total number of blocks
	my $pkey = Crypt::OpenSSL::RSA->new_public_key(read_key_file($keyFile));
	$pkey->use_pkcs1_padding();
	my $outputBlockSize = $pkey->size();
	my $inputBlockSize = $outputBlockSize;#- PADDINGSIZE;
	my $numBlocks = int($inputSize / $inputBlockSize);

	# Start to decrypt.
	my $blockCount = 0;
	my @decryptBuffer = ();

	while ($blockCount < $numBlocks){
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($cryptData, $index, $inputBlockSize);
		$decryptBuffer[$blockCount] = $pkey->public_decrypt($block);
		$blockCount++;
	}


	if ( ($inputSize % $inputBlockSize) != 0 ){
		$numBlocks++;
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($cryptData, $index, $inputBlockSize);
		$decryptBuffer[$blockCount] = $pkey->public_decrypt($block);
	}
	# Return the data built by joining the decrypted blocks
	return join('',@decryptBuffer);
}
=cut

1;

__END__
