# Copyright (c) 2003-2004, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.

package PAPI::PADATH;

use strict;
use PAPI::ApachePoA;
use PAPI::Cookie_Handler;
use Athens::DA;
use Athens::DA::DAA_Packet;
use Athens::DA::DST_Packet;
use URI::Escape;

use Apache2::Const qw(:common);
use APR::URI;
use Apache2::Util qw(:all);


#
# Called to start Athens LAA mode: Request to the AAP
sub reqLAA {
  my $r=shift;
  my $poa=shift;
  my %fields = $poa->stringArgsToHash($r->args());

  my $aid=$poa->{'Athens_IID'};
  my $aap=$poa->{'Athens_AAP'};
  if (!defined $aap || $aap eq "") { $aap = "https://auth.athensams.net/"; }
  my $dah=new Athens::DA($aid,$poa->{'Athens_Key'});
  my $apsa=$poa->{'Athens_psa'};
  if (!defined $apsa || $apsa eq "") { $apsa = "AthensPermissionSet"; }
  my $auia=$poa->{'Athens_uia'};
  if (!defined $auia || $auia eq "") { $auia = "uid"; }
  my $poa_uri=APR::URI->parse($r->pool, $aap);
  my $poa_sch=$poa_uri->scheme();
  my $poa_server=$r->get_server_name();
  my $tsrv=$r->get_server_port();
  if ($tsrv ne "80" && $poa_sch eq "http") { $poa_server.=":$tsrv"; }
  if ($tsrv ne "443" && $poa_sch eq "https") { $poa_server.=":$tsrv"; }
# Build the return URI from Athens AAP
  my $arurl=$poa_sch."://".$poa_server.$poa->{Loc}."/PADATHLAARespHandler";
  $arurl.=$r->uri();
# Get attributes for permission set and user id
  my @atrl=@{$poa->{'attrList'}};
  my $aps="";
  my $auid="";
  for (my $i=0; $i<$#atrl; $i+=2) {
     if ($atrl[$i] eq $apsa) { $aps=$atrl[$i+1]; }
     if ($atrl[$i] eq $auia) { $auid=$atrl[$i+1]; }
  }
  my $daa=new Athens::DA::DAA_Packet(permset=>$aps,ts=>time(),uid=>$auid,
                                     returl=>$arurl);
  my $ap=$dah->create_DA_packet($daa);
# Send redirect to AAP along with access tokens (set by the main handler)
  PAPI::Cookie_Handler::redirect_URL($r,"$aap/?t=daa&id=$aid&p=".uri_escape($ap,'\W'));
  return DONE;
}

#
# Called when a response in Athens LAA mode comes from the AAP
sub respLAA {
  my $r=shift;
  my $poa=shift;
  my $aid=$poa->{'Athens_IID'};
  my %rar = $poa->stringArgsToHash($r->args());
  my $dah=new Athens::DA($aid,$poa->{'Athens_Key'});
  my $attl = $poa->{'Athens_TTL'};
  if (!defined $attl || $attl eq "") { $attl = 60; }
  if ($rar{'t'} eq 'dar') {
    my $rp=$rar{'p'};
    my $dar;
    eval { $dar=$dah->decode_DA_packet($rp); };
    if ($@) { 
      $poa->error("Invalid DAR packet received by PADATH handler");
      return FORBIDDEN;
    }
# Check time stamp from AAP
    if ((time()-$dar->{'ts'})>$attl) {
      $poa->error("Timed-out DAR packet received by PADATH handler");
      return FORBIDDEN;
    }
    my $furi=$r->uri();
    $furi=~/^(.*?)PADATHLAARespHandler(.*?)$/;
    my $ruri=$2.'?'.$r->args;
    my $sr=$r->lookup_uri($ruri);
    my $res=$sr->run();
    if ($res != OK) { 
      $poa->error("respLAA: Exception handling $ruri");
      return $res;
    }
    return DONE;
  }
  else {
    $poa->error("Not a DAR packet received by PADATH LAA handler");
    return FORBIDDEN;
  }
}


#
# Called to answer a request in HDD mode
sub respHDD {
  my $r=shift;
  my $poa=shift;
  my $aap=$poa->{'Athens_AAP'};
  if (!defined $aap || $aap eq "") { $aap = "https://auth.athensams.net/"; }
  my $aid=$poa->{'Athens_IID'};
  my $apsa=$poa->{'Athens_psa'};
  if (!defined $apsa || $apsa eq "") { $apsa = "AthensPermissionSet"; }
  my $auia=$poa->{'Athens_uia'};
  if (!defined $auia || $auia eq "") { $auia = "uid"; 
  my %rar = $poa->stringArgsToHash($r->args());
  my $dah=new Athens::DA($aid,$poa->{'Athens_Key'});
  if ($rar{'t'} eq 'dsr') {
    my $rp=$rar{'p'};
# Get attributes for permission set and user id
    my @atrl=@{$poa->{'attrList'}};
    my $aps="";
    my $auid="";
    for (my $i=0; $i<$#atrl; $i+=2) {
      if ($atrl[$i] eq $apsa) { $aps=$atrl[$i+1]; }
      if ($atrl[$i] eq $auia) { $auid=$atrl[$i+1]; }
    }

    my $dst=new Athens::DA::DST_Packet(permset=>$aps,uid=>$auid,dat=>$rp);
    my $pkt=$dah->create_DA_packet($dst);
    my $rurl="$aap/?t=dst&id=$aid&p=".uri_escape($pkt);
    PAPI::Cookie_Handler::redirect_URL($r,$rurl);
    return DONE;
  }
  else {
    $poa->error("Not a DSR packet received by PADATH HDD handler");
    return FORBIDDEN;
  }
}

1;

__END__
