# Copyright (c) 2001-2007, RedIRIS. All Rights Reserved.
## Parts of the code in this file have been contributed by
# Roman Medina, Telefonica Moviles.
## Parts of the code in this file have been contributed by
#  Luis Melendez, University of Cordoba.
## Parts of the code in this file have been contributed by
#  Maja Gorecka-Wolniewicz, Nicolaus Copernicus University.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.

package PAPI::RewritingProxy;

use strict;
use PAPI::ApachePoA;
use Apache2::Const qw(:common);
use Apache2::RequestRec  ();
use Apache2::RequestUtil ();
use Apache2::Connection();
use Apache2::ServerRec ();
use Apache2::Util qw(:all);
use APR::URI     ();
use Apache2::URI ();
use LWP::UserAgent;
use URI::URL;
use HTML::TokeParser;
use HTML::TreeBuilder;
use HTML::Element;
use HTTP::Cookies;
use HTTP::Request::Common;
use HTTP::Request::Form;
use Data::Dumper;
use CGI;

# Chunked mode buffer
use vars '$acum';

# Variables available for proxy specification
use vars '$papx_hostname';
use vars '$papx_hostdest';
use vars '$papx_pathdest';
use vars '$papx_namedest';
use vars '$papx_location';
use vars '$papx_fulldest';

# Backwards compatibility with older proxy specs
use vars '$host_dest';
use vars '$path_dest';
use vars '$name_dest';

use vars qw($VERSION @ISA @EXPORT);
$| = 1;

# use DynaLoader ();
# @ISA = qw(DynaLoader Exporter);

use Exporter();
@ISA    = qw(Exporter);
@EXPORT = qw(new handler fixLink fetchURL);

sub new {
	my $class = shift;
	my $self  = {};
	bless $self, $class;
	return $self;
}

sub handler {
	my $r = shift;

	$papx_hostname = $r->hostname;
	my $poa = new PAPI::ApachePoA;
	$poa->set_parameters($r);
	$poa->dumpApachePoA("RewritingProxyDump") if ( $poa->{'Debug'} );
	my $urlToFetch = $r->uri;
	my $curl       = $r->construct_url;
	my $puri       = APR::URI->parse( $r->pool, $curl );
	my $sch        = $puri->scheme;
	for my $rqredirect ( @{ $poa->{rqRedirects} } ) {
		my $rde = "";
		my $rdv = "";
		if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
			$rde = eval $rqredirect->{regex};
			if ( !defined $rde ) { $rde = $rqredirect->{regex}; }
			$rdv = eval $rqredirect->{value};
			if ( !defined $rdv ) { $rdv = $rqredirect->{value}; }
		}
		else {
			$rde = $rqredirect->{regex};
			$rdv = $rqredirect->{value};
		}
		my $evs = "\$urlToFetch =~ s($rde)($rdv)gsx";
		eval $evs;
	}

	for my $urlr ( @{ $poa->{RejUrlPatterns} } ) {
		return FORBIDDEN if ( $urlToFetch =~ /$urlr/ );
	}
	for my $urlr ( @{ $poa->{RejPostPatterns} } ) {
		return FORBIDDEN if ( $urlToFetch =~ /$urlr/ );
	}
	$papx_pathdest = $r->location;
	$urlToFetch =~ s/$papx_pathdest//;
	if ( defined $poa->{'Remote_Serv'} ) {
		$papx_hostdest = $poa->{'Remote_Serv'};
		my @thn = ( $papx_hostdest =~ m(^(.*?)://(.*?)\.(.*?)$) );
		$papx_namedest = $thn[1];
	}
	elsif ( defined $poa->{'Remote_Dom'} ) {
		my @thn = ( $urlToFetch =~ m(([^/]+)/(.*?)$) );
		if ( $thn[0] ne "" ) {
			$papx_hostdest =
			  $sch . "://" . $thn[0] . "." . $poa->{'Remote_Dom'};
		}
		else { $papx_hostdest = $sch . "://" . $poa->{'Remote_Dom'}; }
		$papx_namedest = $thn[0];
		$urlToFetch    = "/" . $thn[1];
	}

   # $papx_fulldest is meant to be useful for proxy definitions that can be used
   # in VirtualHost and/or Location contexts.
	$papx_location = ( $papx_pathdest eq "/" ) ? "" : $papx_pathdest;
	$papx_fulldest = "$papx_location/$papx_namedest";
	$papx_fulldest =~ s%/(/+)%/%gg;

	# Keep compatibility with older proxy specs
	$host_dest = $papx_hostdest;
	$name_dest = $papx_namedest;
	$path_dest = $papx_pathdest;

	return &fetchURL( $r, $poa, $urlToFetch );
}

###################################################
# fetchURL
# Returns OK if it's happy or
# an appropriate code if other than 200 is received
sub fetchURL {
	my $r   = shift;
	my $poa = shift;
	my $url = shift;

	# Set local IP address if required by configuration
	if ( defined $poa->{PxIP} && $poa->{PxIP} ne "" ) {
		@LWP::Protocol::http::EXTRA_SOCK_OPTS = ( LocalAddr => $poa->{PxIP} );
	}
	my $ua = new LWP::UserAgent;

	# Let's make sure we send all redirects back to the browser
	$ua->requests_redirectable( [] );
	$ua->parse_head(0);

	# Let's carry the User Agent to the server also.
	my $browser = '';
	if ( defined $r->headers_in->{'User-Agent'} ) {
		$browser = $r->headers_in->{'User-Agent'};
	}
	$ua->agent($browser);

	# Use a proxy server if defined
	if ( defined $poa->{Proxy} && $poa->{Proxy} ne "" ) {
		$ua->proxy( 'http', $poa->{Proxy} );
	}

	# Use HTTP authentication if defined
	my $tauri = URI::URL->new($papx_hostdest);
	for my $hauth ( @{ $poa->{HttpAuth} } ) {
		my ( $br, $bu, $bp ) = ( $hauth->{br}, $hauth->{bu}, $hauth->{bp} );
		if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
			my $brv = eval $br;
			if ( !defined $brv ) { $brv = $br; }
			my $buv = eval $bu;
			if ( !defined $buv ) { $buv = $bu; }
			my $bpv = eval $bp;
			if ( !defined $bpv ) { $bpv = $bp; }
			$ua->credentials( $tauri->host_port(), $brv, $buv, $bpv );
		}
		else { $ua->credentials( $tauri->host_port(), $br, $bu, $bp ); }
	}
	my $rurl = $url;

	# Avoid double slashes into URL (seems to have problems in Windows boxes)
	$rurl =~ s%/(/+)%/%gg;
	my $turl = "";
	if ( defined $poa->{StripLocation} && $poa->{StripLocation} eq "1" ) {
		$turl = $rurl;
	}
	else { $turl = "$papx_pathdest$rurl"; }

	# Chase again double slashes
	$turl =~ s%/(/+)%/%gg;
	$url  = "$papx_hostdest$turl";
	$rurl = $url;

	# We have to append the query string since it got munged by
	# apache when this was first requested.
	$url .= "?" . $r->args if ( $r->args && $url !~ /\?/ );
	my ( $up1, $hostname ) = split( /\/\//, $papx_hostdest );

	my ( $res, $req );
	$poa->accesslog("Fetching remote URL: $url");
	$req = HTTP::Request->new( $r->method, $url );

	# Process and tidy headers:
	# Do not pass to the remote server Lcook, Hcook and PAPI Attributes
	# Rewrite "Host:" to the remote host if present
	# Block all "Connection:" headers
	# Since we need to parse the HTML response, block "Accept-Encoding", so
	# we don't have to uncompress data before parsing it
	my $hdrs = $r->headers_in();
	for my $hdr ( keys %{$hdrs} ) {
		for my $redirect ( @{ $poa->{hdRedirects} } ) {
			if ( $hdr =~ /$redirect->{hdregex}/i ) {
				my $hdv = $hdrs->{$hdr};
				my $rde = "";
				my $rdv = "";
				if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
					$rde = eval $redirect->{regex};
					if ( !defined $rde ) { $rde = $redirect->{regex}; }
					$rdv = eval $redirect->{value};
					if ( !defined $rdv ) { $rdv = $redirect->{value}; }
				}
				else {
					$rde = $redirect->{regex};
					$rdv = $redirect->{value};
				}
				my $evs = "\$hdv =~ s($rde)($rdv)gsx";
				eval $evs;
				$hdrs->{$hdr} = $hdv;
			}
		}
		if ( $hdr eq 'Cookie' ) {
			my $fcv = "";
			for my $ckv ( split /[\s]*;[\s]*/, $hdrs->{$hdr} ) {
				my ( $cn, $cv ) = split /[\s]*=[\s]*/, $ckv, 2;
				if ( $cn !~ /^Hcook$/i && $cn !~ /^Lcook$/i ) {
					if ( $fcv ne "" ) { $fcv .= "; $cn=$cv"; }
					else { $fcv = "$cn=$cv"; }
				}
			}
			$req->header( 'Cookie', $fcv ) if ( $fcv ne "" );
		}
		elsif ( $hdr eq 'Host' ) {
			my ( $phn, $op ) = split /\//, $hostname, 2;
			$req->header( 'Host', $phn );
		}
		elsif ($hdr ne 'Connection'
			&& $hdr ne 'Accept-Encoding'
			&& $hdr !~ /^X-PAPIAttr-/ )
		{
			$req->header( $hdr, $hdrs->{$hdr} );
		}
	}
	if ( $r->method eq 'POST' ) {
		my $len = $r->headers_in->{'Content-length'};
		my $buf;
		$r->read( $buf, $len );
		for my $redirect ( @{ $poa->{postRedirects} } ) {
			my $rde = "";
			my $rdv = "";
			if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
				$rde = eval $redirect->{regex};
				if ( !defined $rde ) { $rde = $redirect->{regex}; }
				$rdv = eval $redirect->{value};
				if ( !defined $rdv ) { $rdv = $redirect->{value}; }
			}
			else {
				$rde = $redirect->{regex};
				$rdv = $redirect->{value};
			}
			my $evs = "\$buf =~ s($rde)($rdv)gsx";
			eval $evs;
		}
		$req->content($buf);
	}
	$ua->{PAPICallBackList} = [ $r, $poa, $url ];
	$acum = undef;
	$res = $ua->request( $req, \&chunkCallBack, $poa->{PxCkSize} );
	if (   ( $res->header('X-PAPICallBackCtl') == 2 )
		|| ( $res->header('X-PAPICallBackCtl') == 3 ) )
	{
		return (OK);
	}
	$res->content($acum) if ($acum);

	# Match form processor with URL without query string
	my $formProc = undef;
	foreach my $u ( keys %{ $poa->{formProc} } ) {
		if ( $url =~ $u ) {
			$formProc = $poa->{formProc}{$u};
			last;
		}
	}
	if (
		defined $formProc
		&& (
			!exists $formProc->{match}
			|| ( exists $formProc->{match}
				&& ( $res->content =~ /$formProc->{match}/ ) )
		)
	  )
	{

		# Get any cookie received with the form
		my @rsc = ();
		$res->scan(
			sub {
				my $hdr = shift;
				my $val = shift;
				if ( $hdr =~ /Set-Cookie/ ) {
					my @cva = split /[\s]*;[\s]*/, $val;
					for my $cv (@cva) {
						if (   $cv !~ /^domain=/i
							&& $cv !~ /^path=/i
							&& $cv !~ /^expires=/i )
						{
							push @rsc, $cv;
						}
					}
				}
			}
		);
		$poa->accesslog("Processing form at $url");
		my $tree     = HTML::TreeBuilder->new_from_content( $res->content );
		my @forms    = $tree->find_by_tag_name('form');
		my $g        = $forms[ $formProc->{formIndex} ];
		my $withname = ( $formProc->{formName} ne "" ) ? 1 : 0;
		my $found    = 0;
		for my $f (@forms) {
			if ( $f->attr('name') eq $formProc->{formName} ) {
				$g     = $f;
				$found = 1;
				last;
			}
		}
		# Add any field defined by Form_Proc and not actually present
		# Useful for dealing with Javascript managed forms
		for my $f (keys %{$formProc->{settings}}) {
        	my $field = $g->look_down("name",$f);
        	if($field == undef) {
          		my $h = HTML::Element->new('input', 'name'=>$f, 'type'=>'hidden' );
          		$g->push_content($h);
          		$poa->{'log'}->debug( "Inserting field " . $f ) if ( $poa->{'Debug'} );
        	}
      	}

		if ( ( $withname && $found ) || !$withname ) {
			$poa->{'log'}->debug("The URL for processing the FORM: $url;")
			  if ( $poa->{'Debug'} );
			my $thef = HTTP::Request::Form->new( $g, $url );
			$poa->{'log'}->debug( "The FORM base: " . $thef->base() )
			  if ( $poa->{'Debug'} );
			$poa->{'log'}->debug( "The FORM link: " . $thef->link() )
			  if ( $poa->{'Debug'} );
			my @buttons = $thef->buttons();
			my $button;

			# When Form_Processor has a field named button this field value
			# is treated as the name of buttons
			my $cont = 0;
			if (   defined $formProc->{buttonName}
				|| defined $formProc->{buttonIndex} )
			{
				if ( defined $formProc->{buttonName} ) {
					$button = $formProc->{buttonName};
				}
				elsif ( defined $formProc->{buttonIndex} ) {
					$button = $buttons[ $formProc->{buttonIndex} ];
				}
				foreach my $fb (@buttons) {
					if ( $fb eq $button ) { $cont = 1; last; }
				}
			}
			else {
				$cont   = 1;
				$button = $buttons[0];
			}
			if ($cont) {
				$poa->accesslog( "Sending form data to " . $thef->link );
				for my $f ( keys %{ $formProc->{settings} } ) {
					my $fv = undef;
					if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
						$fv = eval $formProc->{settings}{$f};
					}
					if ( !defined $fv ) {
						$fv = $formProc->{settings}{$f};
					}
					$thef->field( $f, $fv );
				}
				$poa->accesslog("Pressing the button: $button");
				my $freq = $thef->press($button);

	  # Send along with the new request the cookies received from the client
	  # in the original request and those received from the remote server in the
	  # initial response
				for my $hdr ( keys %{$hdrs} ) {
					if ( $hdr eq 'Cookie' ) {
						my $fcv = "";
						for my $ckv ( split /[\s]*;[\s]*/, $hdrs->{$hdr} ) {
							my ( $cn, $cv ) = split /[\s]*=[\s]*/, $ckv, 2;
							if ( $cn !~ /Hcook/i && $cn !~ /Lcook/i ) {
								if ( $fcv ne "" ) {
									$fcv .= "; $cn=$cv";
								}
								else {
									$fcv = "$cn=$cv";
								}
							}
						}
						$freq->header( 'Cookie', $fcv ) if ( $fcv ne "" );
					}
				}
				for my $cv (@rsc) {
					$freq->header( 'Cookie', $cv );
				}
				$ua->{PAPICallBackList} = [ $r, $poa, $url ];
				$acum = undef;
				$res = $ua->request( $freq, \&chunkCallBack, $poa->{PxCkSize} );
				if (   ( $res->header('X-PAPICallBackCtl') == 2 )
					|| ( $res->header('X-PAPICallBackCtl') == 3 ) )
				{
					return (OK);
				}
				$res->content($acum) if ($acum);
			}
			else {
				$poa->error(
"Button $button does not exist in this form, sending form data to "
					  . $thef->link
					  . " abandoned" );
			}
		}
	}
	$r->status( $res->code );
	procHeaders( $res, $poa, $r, $url );
	my $outString = "";

	#
	# Parse and rewrite HTML content
	if ( $res->header('Content-type') =~ /html/i ) {
		my $content  = $res->content;
		my $baseHref = "";
		my $inScript = 0;
		my $p        = HTML::TokeParser->new( \$content )
		  || $poa->error("No content available from remote URL $url: $!");

		# Avoid entity decoding inside attributes
		$p->attr_encoded(1);

		# Unless No_XML is set to 1, allow XHTML tags (<x/>)
		if ( ( defined $poa->{NoXML} && $poa->{NoXML} eq "1" ) ) {
			$p->xml_mode(0);
		}
		else { $p->xml_mode(1); }
		while ( my $tolkens = $p->get_token ) {
			if (   $tolkens->[0] eq 'T'
				|| $tolkens->[0] eq 'C'
				|| $tolkens->[0] eq 'D' )
			{
				if ( $inScript
					|| ( defined $poa->{RwAll} && $poa->{RwAll} eq "1" ) )
				{
					for my $redirect ( @{ $poa->{redirects} } ) {
						my $rde = "";
						my $rdv = "";
						if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" )
						  )
						{
							$rde = eval $redirect->{regex};
							if ( !defined $rde ) { $rde = $redirect->{regex}; }
							$rdv = eval $redirect->{value};
							if ( !defined $rdv ) { $rdv = $redirect->{value}; }
						}
						else {
							$rde = $redirect->{regex};
							$rdv = $redirect->{value};
						}
						my $evs = "\$tolkens->[1] =~ s($rde)($rdv)gsx";
						eval $evs;
					}
					for my $redirect ( @{ $poa->{url_redirects} } ) {
						my $rde = "";
						my $rdv = "";
						if ( $url =~ $redirect->{urlregex} ) {
							if (
								(
									defined $poa->{EvalPx}
									&& $poa->{EvalPx} eq "1"
								)
							  )
							{
								$rde = eval $redirect->{regex};
								if ( !defined $rde ) {
									$rde = $redirect->{regex};
								}
								$rdv = eval $redirect->{value};
								if ( !defined $rdv ) {
									$rdv = $redirect->{value};
								}
							}
							else {
								$rde = $redirect->{regex};
								$rdv = $redirect->{value};
							}
							my $evs = "\$tolkens[1] =~ s($rde)($rdv)gsx";
							eval $evs;
						}
					}
				}
				$outString .= $tolkens->[1];
			}
			elsif ( $tolkens->[0] eq 'E' ) {

				# End script rewriting
				if (   $tolkens->[1] =~ /^script$/i
					|| $tolkens->[1] =~ /^style$/i )
				{
					$inScript = 0;
				}

 # Only write tag out if we are in a real end token (XML mode forces this check)
				$outString .= "</" . $tolkens->[1] . ">"
				  if ( $tolkens->[2] ne "" );
			}
			elsif ( $tolkens->[0] eq 'S' ) {

	   # Convert all attribute names to lowercase, so the references inside work
	   # with XML mode on
				for my $ki ( keys %{ $tolkens->[2] } ) {
					my $kj = lc $ki;
					if ( $ki ne $kj ) {
						$tolkens->[2]{$kj} = $tolkens->[2]{$ki};
						delete $tolkens->[2]{$ki};
					}
				}

		   # End script rewriting. It will be activated if starting a script tag
				$inScript = 0;
				my $text    = $tolkens->[4];
				my @oldLink = ();
				my @newLink = ();
				if ( $tolkens->[1] =~ /^a$/i ) {
					if ( $tolkens->[2]{href} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{href}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{href} );
					}
				}
				elsif ( $tolkens->[1] =~ /^base$/i ) {
					if ( $tolkens->[2]{href} ) {
						push @newLink,
						  &fixLink( $r, $poa, $tolkens->[2]{href}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{href} );
						$url = $tolkens->[2]{href};
					}
				}
				elsif ( $tolkens->[1] =~ /^meta$/i ) {
					if ( $tolkens->[2]{content} =~ /url/i ) {
						my ( $junk, $tmpLink ) =
						  split( /=/, $tolkens->[2]{content} );
						push @newLink, &fixLink( $r, $poa, $tmpLink, $url );
						push @oldLink, regexEscape($tmpLink);
					}
				}
				elsif ( $tolkens->[1] =~ /^(area|link)$/i ) {
					if ( $tolkens->[2]{href} ) {
						push @newLink,
						  &fixLink( $r, $poa, $tolkens->[2]{href}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{href} );
					}
				}
				elsif ( $tolkens->[1] =~ /^(frame|img|iframe)$/i ) {
					if ( $tolkens->[2]{src} ) {
						push @newLink,
						  &fixLink( $r, $poa, $tolkens->[2]{src}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{src} );
					}
				}
				elsif ( $tolkens->[1] =~ /^form$/i ) {
					if ( $tolkens->[2]{action} ) {
						push @newLink,
						  &fixLink( $r, $poa, $tolkens->[2]{action}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{action} );
					}
				}
				elsif ( $tolkens->[1] =~ /^(table|td|body)$/i ) {
					if ( $tolkens->[2]{background} ) {
						push @newLink,
						  &fixLink( $r, $poa, $tolkens->[2]{background}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{background} );
					}
				}

		  # Activate rewriting for the content of style specs (imported URLs...)
				elsif ( $tolkens->[1] =~ /^style$/i ) { $inScript = 1; }
				elsif ( $tolkens->[1] =~ /^script$/i ) {
					if ( $tolkens->[2]{src} ) {
						push @newLink,
						  &fixLink( $r, $poa, $tolkens->[2]{src}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{src} );
					}

				  # Inline script (not from a source object). Activate rewriting
					else { $inScript = 1; }
				}
				elsif ( $tolkens->[1] =~ /^option$/i ) {
					if ( $tolkens->[2]{value} ) {
						push @newLink,
						  &fixLink( $r, $poa, $tolkens->[2]{value}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{value} );
					}
				}
				elsif ( $tolkens->[1] =~ /^input$/i ) {
					if ( $tolkens->[2]{value} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{value}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{value} );
					}
					if ( $tolkens->[2]{src} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{src}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{src} );
					}
				}
				elsif ( $tolkens->[1] =~ /^object$/i ) {
					if ( $tolkens->[2]{archive} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{archive}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{archive} );
					}
					if ( $tolkens->[2]{codebase} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{codebase}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{codebase} );
					}
					if ( $tolkens->[2]{data} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{data}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{data} );
					}
					if ( $tolkens->[2]{map} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{map}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{map} );
					}
				}
				elsif ( $tolkens->[1] =~ /^param$/i ) {
					if ( $tolkens->[2]{value} ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{value}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{value} );
					}
				}

				# Process script and style attributes
				for my $att ( keys %{ $tolkens->[2] } ) {
					$att = lc $att;
					if ( $att =~ /^on/i ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{$att}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{$att} );
					}
					elsif ( $att =~ /^style$/i ) {
						push @newLink,
						  &fixAnchor( $r, $poa, $tolkens->[2]{$att}, $url );
						push @oldLink, regexEscape( $tolkens->[2]{$att} );
					}
				}

		 # Some servers send character entities inside URLs, that make the regex
		 # substiting $oldLink by $newLink fail. Up to now, only '&amp;' -> '&'
		 # is considered.
		 # Not longer needed in the current parser configuration
		 #        if ($#newLink >= 0) { $text =~ s(\&amp;)(&)gs; }
				for my $i ( 0 .. $#newLink ) {
					$text =~ s($oldLink[$i])($newLink[$i])gs;
				}
				$outString .= $text;
			}
			else { $outString .= $tolkens->[4]; }
		}
	}

	#
	# Non-HTML content. Check for MIME types and URL patterns
	# where rewriting is applicable
	else {
		$outString = $res->content;
		for my $redirect ( @{ $poa->{redirects} } ) {
			my $rde = "";
			my $rdv = "";
			if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
				$rde = eval $redirect->{regex};
				if ( !defined $rde ) { $rde = $redirect->{regex}; }
				$rdv = eval $redirect->{value};
				if ( !defined $rdv ) { $rdv = $redirect->{value}; }
			}
			else {
				$rde = $redirect->{regex};
				$rdv = $redirect->{value};
			}
			my $evs = "\$outString =~ s($rde)($rdv)gsx";
			eval $evs;
		}
		for my $redirect ( @{ $poa->{url_redirects} } ) {
			my $rde = "";
			my $rdv = "";
			if ( $url =~ $redirect->{urlregex} ) {
				if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
					$rde = eval $redirect->{regex};
					if ( !defined $rde ) { $rde = $redirect->{regex}; }
					$rdv = eval $redirect->{value};
					if ( !defined $rdv ) { $rdv = $redirect->{value}; }
				}
				else {
					$rde = $redirect->{regex};
					$rdv = $redirect->{value};
				}
				my $evs = "\$outString =~ s($rde)($rdv)gsx";
				eval $evs;
			}
		}
	}
	if ( !defined( $res->header('Transfer-encoding') ) ) {
		$r->headers_out->set( 'Content-length' => length($outString) );
	}

	# Some browsers are reported to be extremely sensitive to content types
	# containing extra blank characters. Clean it before sending
	my $ttc = $res->header('Content-type');
	$ttc =~ s/(.*?)[\s]+$/$1/;
	$r->content_type($ttc);
	$r->print($outString);
	return (OK);
}

#################################
# chunkCallBack
# Manage chunked transfer of data
sub chunkCallBack {
	my ( $data, $response, $protocol ) = @_;
	my $cont = $response->header('X-PAPICallBackCtl');
	my ( $r, $poa, $url ) = @{ $protocol->{ua}->{PAPICallBackList} };
	if ( !$cont ) {
		$cont = 2;
		if ( $response->header('Content-type') =~ /html/i ) {
			$response->header( 'X-PAPICallBackCtl', 1 );
			$cont = 1;
		}
		elsif ( $response->header('Content-type') =~ /$poa->{RewMimeTypes}/i ) {
			$response->header( 'X-PAPICallBackCtl', 1 );
			$cont = 1;
		}
		else {
			for my $patt ( @{ $poa->{RewUrlPatterns} } ) {
				if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
					my $rpt = eval $patt;
					if ( !defined $rpt ) { $rpt = $patt; }
					if ( $url =~ /$rpt/ ) { $cont = 1; last; }
				}
				else {
					if ( $url =~ /$patt/ ) { $cont = 1; last; }
				}
			}
		}
		$response->header( 'X-PAPICallBackCtl', $cont );
		if ( $cont == 2 ) {
			$r->status( $response->code );
			procHeaders( $response, $poa, $r, $url );

		 # Some browsers are reported to be extremely sensitive to content types
		 # containing extra blank characters. Clean it before sending
			my $ttc = $response->header('Content-type');
			$ttc =~ s/(.*?)[\s]+$/$1/;
			$r->content_type($ttc);
			if (defined $response->header('Content-length')) {
				$r->headers_out->set(
					'Content-length' => $response->header('Content-length') );
			}
		}
	}
	if ( $cont == 1 ) { $acum .= $data; }
	else {
		my $c = $r->connection;
		if ( $c->aborted ) {
			$response->header( 'X-PAPICallBackCtl', 3 );
			$poa->error("Chunked transfer aborted");
			die("Chunked transfer aborted. Connection closed");
		}
		$r->print($data);
	}
}

############################################################################
# procHeaders
# Process and tidy received headers:
# Cookies are transformed so they can be sent back to the originating server
# Redirects (and the non-standard Refresh:) are adapted to the rewriten URLs
sub procHeaders {
	my $res = shift;
	my $poa = shift;
	my $r   = shift;
	my $url = shift;

	$res->scan(
		sub {
			my $hdr = shift;
			my $val = shift;
			if ( $hdr =~ /Set-Cookie/i ) {
				my $cfv = undef;
				my @cva = split /[\s]*;[\s]*/, $val;
				if ( !defined $poa->{'Remote_Dom'} ) {
					for my $cv (@cva) {
						if ( $cv !~ /^domain=/i ) {
							if ( !defined $cfv ) { $cfv = $cv; }
							else { $cfv .= "; $cv"; }
						}
					}
				}
				else {
					my %ch = ();
					for my $cv (@cva) {
						my ( $pc, $vc ) = split /[\s]*=[\s]*/, $cv;
						$ch{ lc($pc) } = $vc;
					}
					for my $cv (@cva) {
						if ( $cv !~ /^domain=/i ) {
							if ( $cv =~ /^path=/i ) {
								if ( !exists $ch{"domain"} ) {
									$cv =
									  "path=/" . $papx_namedest . $ch{"path"};
								}
							}
							if ( !defined $cfv ) { $cfv = $cv; }
							else { $cfv .= "; $cv"; }
						}
					}
				}
				$val = $cfv;
				$r->err_headers_out->add( $hdr => $val );
			}
			elsif ( $hdr =~ /Location/i ) {
				$val = &fixLink( $r, $poa, $val, $url );
				$r->headers_out->set( $hdr => $val );
			}
			elsif ( $hdr =~ /Refresh/i ) {
				my @rva = split /[\s]*;[\s]*/, $val;
				my ( $ui, @rv ) = split /[\s]*=[\s]*/, $rva[1];
				my $rvv = join "=", @rv;
				$rva[1] = "URL=" . &fixLink( $r, $poa, $rvv, $url );
				$val = join ";", @rva;
				$r->headers_out->set( $hdr => $val );
			}
			else {
				$r->headers_out->set( $hdr => $val )
				  unless ( $hdr =~ /X-PAPICallBackCtl/ );
			}
		}
	);
}

##################################################################
# fixAnchor
# This is a wrapper to fixLink for dealing with JavaScript anchors
sub fixAnchor {
	my $r        = shift;
	my $poa      = shift;
	my $link     = shift;
	my $baseLink = shift;

	my $rtv = &fixLink( $r, $poa, $link, $baseLink );

	# If a special method is in use, apply redirects
	# At the moment, only "javascript:" is considered
	if ( $rtv =~ /^javascript:/i ) {
		for my $redirect ( @{ $poa->{redirects} } ) {
			my $rde = "";
			my $rdv = "";
			if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
				$rde = eval $redirect->{regex};
				if ( !defined $rde ) { $rde = $redirect->{regex}; }
				$rdv = eval $redirect->{value};
				if ( !defined $rdv ) { $rdv = $redirect->{value}; }
			}
			else {
				$rde = $redirect->{regex};
				$rdv = $redirect->{value};
			}
			my $evs = "\$rtv =~ s($rde)($rdv)gsx";
			eval $evs;
		}
		for my $redirect ( @{ $poa->{url_redirects} } ) {
			my $rde = "";
			my $rdv = "";
			if ( $baseLink =~ $redirect->{urlregex} ) {
				if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
					$rde = eval $redirect->{regex};
					if ( !defined $rde ) { $rde = $redirect->{regex}; }
					$rdv = eval $redirect->{value};
					if ( !defined $rdv ) { $rdv = $redirect->{value}; }
				}
				else {
					$rde = $redirect->{regex};
					$rdv = $redirect->{value};
				}
				my $evs = "\$rtv =~ s($rde)($rdv)gsx";
				eval $evs;
			}
		}
	}
	return $rtv;
}

###########################################################
# fixLink
# Gets a reference to a remote resource. It can be a URL or
# a relative reference. fixLink will apply the rewriting
# rules to it, returning the fixed reference
sub fixLink {
	my $r           = shift;
	my $poa         = shift;
	my $link        = shift;
	my $baseLink    = shift;
	my $server_name = $r->get_server_name;
	my $i;
	my $urlPath     = "";
	my $hostName    = "";
	my $base_name   = $r->location();
	my $tdn         = undef;
	my $server_port = $r->get_server_port();
	my $curl        = $r->construct_url;
	my $puri        = APR::URI->parse( $r->pool, $curl );

	my $sch = $puri->scheme;
	my ( $protocol, $hostname, @dirs );

	## Clean double slash //
	$link     =~ s/([^:])\/+/$1\//g;
	$baseLink =~ s/([^:])\/+/$1\//g;

	( $protocol, $hostname ) = split( /\/\//, $baseLink );
	( $hostname, @dirs )     = split( /\//,   $hostname );
	if ( $dirs[1] =~ /\w/ ) { $urlPath = join( "/", @dirs ); }
	else { $urlPath = $dirs[0] unless ( $baseLink =~ /\w$/ ); }
	if ( $baseLink =~ /\w$/ && $dirs[1] =~ /\w/ ) {
		for ( $i = length($urlPath) ; $i > 0 ; $i-- ) {
			last if ( substr( $urlPath, $i, 1 ) eq "/" );
		}
		$urlPath = substr( $urlPath, 0, $i ) if ( $i && $urlPath =~ /\// );
	}

	if ( defined $poa->{'Remote_Dom'} ) {
		$tdn = $poa->{'Remote_Dom'};
		my ($thn) = ( $hostname =~ m(^(.+?)\.($tdn)(:\d+)?$) );
		$base_name .= "/" . $thn;
	}
	$hostname = "$protocol//$hostname";
	if (   ( $sch eq 'http' && $server_port != 80 )
		|| ( $sch eq 'https' && $server_port != 443 ) )
	{
		$server_name .= ":" . $server_port;
	}

	if ( $link =~ /^\// ) {
		$link = $sch . "://" . $server_name . $base_name . $link;
	}
	elsif ( $link =~ /^\.\.$/ || $link =~ /^\.\.\// ) {
		$link =
		    $sch . "://"
		  . $server_name
		  . $base_name . "/"
		  . $urlPath . "/"
		  . $link;
	}
	elsif ( !defined $tdn && $link =~ /^$papx_hostdest/i ) {
		$link =~ s/^$papx_hostdest//i;
		$link = $sch . "://" . $server_name . $base_name . $link;
	}
	elsif ( defined $tdn && $link =~ /^\w+?\.$tdn/i ) {
		my ( $up1, $up2 ) = split( /\/\//, $link );
		my ($thn) = ( $up2 =~ m(^(.+?)\.($tdn)) );
		$link =~ s/^(.+?)\.$tdn//i;
		$link = $sch . "://" . $server_name . "/" . $thn . $link;
	}

	for my $redirect ( @{ $poa->{redirects} } ) {
		my $rde = "";
		my $rdv = "";
		if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
			$rde = eval $redirect->{regex};
			if ( !defined $rde ) { $rde = $redirect->{regex}; }
			$rdv = eval $redirect->{value};
			if ( !defined $rdv ) { $rdv = $redirect->{value}; }
		}
		else {
			$rde = $redirect->{regex};
			$rdv = $redirect->{value};
		}
		my $evs = "\$link =~ s($rde)($rdv)gsx";
		eval $evs;
	}
	for my $redirect ( @{ $poa->{url_redirects} } ) {
		my $rde = "";
		my $rdv = "";
		if ( $baseLink =~ $redirect->{urlregex} ) {
			if ( ( defined $poa->{EvalPx} && $poa->{EvalPx} eq "1" ) ) {
				$rde = eval $redirect->{regex};
				if ( !defined $rde ) { $rde = $redirect->{regex}; }
				$rdv = eval $redirect->{value};
				if ( !defined $rdv ) { $rdv = $redirect->{value}; }
			}
			else {
				$rde = $redirect->{regex};
				$rdv = $redirect->{value};
			}
			my $evs = "\$link =~ s($rde)($rdv)gsx";
			eval $evs;
		}
	}
	## Clean double slash //
	$link =~ s/([^:])\/+/$1\//g;
	return $link;
}

############################################################
# regexEscape
# This silly little regex fixes \-[]()*+?|&#. in the URL.
# Without this regex, any of these characters in a URL will
# cause a server error (unless they resolve into something
# sensible to regex, in which case the server does something
# magical and unpredictable with the URL)
sub regexEscape {
	my $url = shift;
	$url =~ s/(\\|\-|\[|\]|\(|\)|\*|\+|\?|\||\&|\#|\.)/\\$1/g;
	return $url;
}

=pod

############################################################
# debugRegex
# This function sends a log message to a file for debug purpouse
# what: What have been rewite
# original: The Original data to be rewite
# rewrote: The resultant data of rewriting
# regex: The REGEX used to trewrite the data
#
# 

sub debugRegex {
  my $what = shift;
  my $original = shift;
  my $rewrote = shift;
  my $regex = shift;

  my $outString = '';
  my $currentTime = '';

	
	my $currentTime = gmtime();
	
	$outStrign = "[".$currentTime."] ".$msg."\n";
	print RPLF, $outString;
}

############################################################
# debugRegex
# This function sends a log message to a file for debug purpouse
#
# mesg: The mesage to be logged
sub debug {
	my $msg = shift;
	
	my $currentTime = gmtime();
	
	$outStrign = "[".$currentTime."] ".$msg."\n";
	print RPLF, $outString;

}


############################################################
# headerDebug
# This function make the header for the debug file
sub headerDebug {
	my $localURL = shift;
	
	my $outString = "############################################################\n";
	$outString   .= "#                                                           \n";
	$outString   .= "# URL to fetch:                                             \n";
	$outString   .= "#     $localURL\n";
	$outString   .= "#                                                           \n";
	$outString   .= "############################################################\n";
	
	print RPLF, $outString;
	
}
=cut

1;

__END__
